package polyglot.ext.ejava.ast;

import java.util.*;

import polyglot.ast.JL;
import polyglot.ast.Node;
import polyglot.ext.ejava.types.EJavaContext;
import polyglot.ext.ejava.types.EJavaTypeSystem;
import polyglot.ext.ejava.types.ExtensionType;
import polyglot.ext.jl.ast.JL_c;
import polyglot.ext.jl5.ast.JL5MethodDecl;
import polyglot.ext.jl5.types.JL5MethodInstance;
import polyglot.types.MethodInstance;
import polyglot.types.Named;
import polyglot.types.ReferenceType;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.visit.TypeChecker;

public class EJavaMethodDecl_Del extends JL_c implements JL {

    @Override
    public Node typeCheck(TypeChecker tc) throws SemanticException {
        JL5MethodDecl methodDecl = (JL5MethodDecl) node();
        EJavaContext ctx = (EJavaContext) tc.context();
        ExtensionType et = ctx.currentExtension();
        if (et == null)
            return methodDecl.typeCheck(tc);

        if (methodDecl.flags().isAbstract())
            throw new SemanticException("extensions cannot declare abstract methods");
        if (methodDecl.flags().isProtected())
            throw new SemanticException("extensions cannot declare protected methods");
        if (methodDecl.flags().isStatic())
            throw new SemanticException("extensions cannot declare static methods");
        if (!et.isTop() && methodDecl.flags().isFinal())
            throw new SemanticException("final methods can only be declared in top extensions");

        JL5MethodInstance mi = (JL5MethodInstance) methodDecl.methodInstance();
        if (et.isTop() && mi.overrides().size() > 1) {
            for (Iterator i = mi.overrides().iterator(); i.hasNext();) {
                ReferenceType t = ((MethodInstance) i.next()).container();
                if (!et.isImplicitCastValid(t))
                    throw new SemanticException("extensions cannot override methods of their base class / interface");
            }
        }

        ExtensionType top = null;
        try {
            Named n = ctx.importTable().find(et.extensionName());
            if (n instanceof ExtensionType) {
                top = (ExtensionType) n;
            } else {
                throw new SemanticException(et.extensionName() + " is not an extension type");
            }
        } catch (SemanticException e) {
            throw new SemanticException("extension type " + et.extensionName() + " not found");
        }

        if (!et.isTop() && !(tc.typeSystem().hasMethod(top, mi) || tc.typeSystem().hasMethod(top, mi.erasure()))) {
            throw new SemanticException("the non-private method " + mi + " is not declared in the top extension implementation of "
                    + et.name());
        }

        if (!et.isTop()) {
            EJavaTypeSystem ts = (EJavaTypeSystem) tc.typeSystem();
            List<Type> typeargs = new ArrayList<Type>(mi.typeVariables());
            JL5MethodInstance mj = ts.findJL5Method(top, mi.name(), mi.formalTypes(), typeargs, top);
            if (tc.typeSystem().isAccessible(mj, tc.context()))
                checkExtensionOverride(mi, mj);
        }
        return methodDecl.typeCheck(tc);
    }

    protected void checkExtensionOverride(MethodInstance mCurr, MethodInstance mTop) throws SemanticException {

        if (mTop.flags().isFinal())
            throw new SemanticException("the method " + mCurr + " overrides a final method");

        if (!mTop.flags().equals(mCurr.flags()))
            throw new SemanticException("the methods " + mCurr + " and " + mTop + " have different flags");

        if (!mTop.returnType().equals(mCurr.returnType()))
            throw new SemanticException("the methods " + mCurr + " and " + mTop + " have different return types");

        Set topThrows = new HashSet(mTop.throwTypes());
        Set currThrows = new HashSet(mCurr.throwTypes());
        if (!topThrows.equals(currThrows))
            throw new SemanticException("the methods " + mCurr + " and " + mTop + " have different return types");
    }

}
