package polyglot.ext.ejava.ast;

import java.util.List;

import polyglot.ast.Expr;
import polyglot.ast.Node;
import polyglot.ast.Term;
import polyglot.ast.TypeNode;
import polyglot.ext.ejava.types.ExtensionType;
import polyglot.ext.jl.ast.Expr_c;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.util.CodeWriter;
import polyglot.util.Position;
import polyglot.visit.*;

public class ExtensionApplication_c extends Expr_c implements ExtensionApplication {
    protected Expr expression;

    protected TypeNode extension;

    protected boolean augmented;

    public ExtensionApplication_c(Position pos, Expr expression, TypeNode extension,
            boolean augmented) {
        super(pos);
        this.expression = expression;
        this.extension = extension;
        this.augmented = augmented;
    }

    public void augmented(boolean a) {
        augmented = a;
    }

    public Expr expression() {
        return expression;
    }

    public ExtensionApplication expression(Expr expression) {
        ExtensionApplication_c r = (ExtensionApplication_c) copy();
        r.expression = expression;
        return r;
    }

    public TypeNode extension() {
        return extension;
    }

    public ExtensionApplication extension(TypeNode extension) {
        ExtensionApplication_c r = (ExtensionApplication_c) copy();
        r.extension = extension;
        return r;
    }

    public void prettyPrint(CodeWriter w, PrettyPrinter pp) {
        ExtensionType et = (ExtensionType) extension().type();
        String name = (augmented ? expression.position().file().split("\\.")[0] + "$" : "")
                + et.name();
        if (et.isLocal())
            w.write("new " + name + "()");
        else
            w.write(name);
        w.write(".expand(");
        expression().prettyPrint(w, pp);
        w.write(")");
    }

    public Node typeCheck(TypeChecker tc) throws SemanticException {
        Type t = extension().type();
        if (!(t instanceof ExtensionType))
            throw new SemanticException("this is not an extension type", extension().position());
        ExtensionType et = (ExtensionType) t;
        et = (ExtensionType) tc.context().find(et.name());
        // System.out.println("et = " + et + ", istop="+et.isTop()+",
        // parent="+et.parent());
        if (!tc.typeSystem().isSubtype(expression.type(), et.superType()))
            throw new SemanticException("extension " + et.name() + " does not " + "apply to type "
                    + expression().type(), position());
        ExtensionApplication r = this;
        /*
         * I DON'T KNOW WHAT THIS WAS, AND IT WAS COMMENTED OUT. PROBABLY SAFE
         * TO REMOVE. if (tc.context().currentClass() instanceof ExtensionType &&
         * expression() instanceof Special &&
         * ((Special)expression()).kind().equals(Special.THIS)) { Expr e =
         * tc.nodeFactory().Super(expression().position()).type(tc.context().currentClass().superType());
         * r = r.expression(e); }
         */
        // THIS IS JUST WRONG, IT WAS THERE BEFORE TOO
        // return r.type((Type)et.interfaces().get(0));
        // I THINK THIS MAKES SENSE:
        return type(et.interfaceType());
    }

    public Type childExpectedType(Expr child, AscriptionVisitor av) {
        if (child == expression)
            return ((ExtensionType) extension().type()).superType();
        else
            return child.type();
    }

    public Node visitChildren(NodeVisitor v) {
        return expression((Expr) visitChild(expression, v)).extension(
                (TypeNode) visitChild(extension, v));
    }

    public Term entry() {
        return expression.entry();
    }

    public List acceptCFG(CFGBuilder v, List succs) {
        v.visitCFG(expression, this);
        return succs;
    }

    public String toString() {
        return "(" + expression() + " with " + extension() + ")";
    }
}
