package polyglot.ext.ejava.types;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import polyglot.ext.jl5.types.JL5Context_c;
import polyglot.types.*;

public class EJavaContext_c extends JL5Context_c implements EJavaContext {

    public EJavaContext_c(TypeSystem ts) {
        super(ts);
    }

    List<ExtensionType> extensions = new ArrayList<ExtensionType>();

    public void addExtension(ExtensionType e) {
        if (extensions == null)
            extensions = new ArrayList<ExtensionType>();
        extensions.add(e);
    }

    public List<ExtensionType> extensions() {
        return extensions;
    }

    public Context extensions(List<ExtensionType> ee) {
        EJavaContext_c r = (EJavaContext_c) copy();
        r.extensions = ee;
        return r;
    }

    public String toString() {
        return super.toString() + ", extensions=" + extensions;
    }

    /*
     * This is from JL's Context_c. I may need to tweak it to get extension
     * fields working (similar to what I did for findMethodScope, probably)
     * 
     * public ClassType findFieldScope(String name) throws SemanticException {
     * if (Report.should_report(TOPICS, 3)) Report.report(3, "find-field-scope " +
     * name + " in " + this);
     * 
     * VarInstance vi = findVariableInThisScope(name);
     * 
     * if (vi instanceof FieldInstance) { if (Report.should_report(TOPICS, 3))
     * Report.report(3, "find-field-scope " + name + " in " + vi); return type; }
     * 
     * if (vi == null && outer != null) { return outer.findFieldScope(name); }
     * 
     * throw new SemanticException("Field " + name + " not found."); }
     */

    public ClassType findMethodScope(String name) throws SemanticException {
        if (this.currentClass() != null && ts.hasMethodNamed(this.currentClass(), name))
            return this.currentClass();
        if (extensions() != null) {
            for (Iterator i = extensions().iterator(); i.hasNext();) {
                ExtensionType et = (ExtensionType) i.next();
                if (!et.applicableTo(currentClass())) 
                    continue;
                if (ts.hasMethodNamed(et, name))
                    return currentClass();
            }
        }
        if (outer != null)
            return outer.findMethodScope(name);
        throw new SemanticException("Method " + name + " not found.");
    }

    /*
     * This is from JL's Context_c. I may need to tweak it to get extension
     * fields working (similar to what I did for findMethod, probably)
     * 
     * public FieldInstance findField(String name) throws SemanticException {
     * VarInstance vi = findVariableSilent(name);
     * 
     * if (vi instanceof FieldInstance) { FieldInstance fi = (FieldInstance) vi;
     * 
     * if (! ts.isAccessible(fi, this)) { throw new SemanticException("Field " +
     * name + " not accessible."); }
     * 
     * if (Report.should_report(TOPICS, 3)) Report.report(3, "find-field " +
     * name + " -> " + fi); return fi; }
     * 
     * throw new NoMemberException(NoMemberException.FIELD, "Field " + name + "
     * not found."); }
     */

    public MethodInstance findMethod(String name, List argTypes) throws SemanticException {
        // Check for any method with the appropriate name.
        // If found, stop the search since it shadows any enclosing
        // classes method of the same name.
        if (currentClass() != null) {
            boolean ok = ts.hasMethodNamed(currentClass(), name);
            EJavaContext c = this;
            while (!ok && c != null) {
                List ee = c.extensions();
                if (ee == null)
                    ee = new LinkedList();
                Iterator i = ee.iterator();
                while (!ok && i.hasNext()) {
                    ExtensionType et = (ExtensionType) i.next();
                    if (ts.hasMethodNamed(et, name)) {
                        ok = true;
                        break;
                    }
                }
                c = (EJavaContext) c.pop();
            }
            // Found a class which has a method of the right name.
            // Now need to check if the method is of the correct type.
            if (ok)
                return ((EJavaTypeSystem) ts).extensionFriendlyFindMethod(currentClass(), name,
                        argTypes, this);
        }
        if (outer != null)
            return outer.findMethod(name, argTypes);
        throw new SemanticException("Method " + name + " not found.");
    }

    public ExtensionType currentExtension() {
        ClassType c = currentClass();
        if (c instanceof ExtensionType) {
            return (ExtensionType) c;

        }
        else {
            return null;
        }
    }

    public Context pushExtensions(List<ExtensionType> ee) {
        EJavaContext_c r = (EJavaContext_c) push();
        r.extensions = ee;
        return r;
    }


}
