package polyglot.ext.ejava.types;

import java.util.HashSet;
import java.util.Set;

import polyglot.ext.jl.types.ParsedClassType_c;
import polyglot.frontend.Source;
import polyglot.types.LazyClassInitializer;
import polyglot.types.ParsedClassType;
import polyglot.types.Type;

public class ExtensionType_c extends ParsedClassType_c implements ExtensionType {

    protected boolean okAsType;
    protected String extensionName;
    protected boolean isTop;
    protected Set cases;
    protected ExtensionType parent;

    public ExtensionType_c(EJavaTypeSystem ts, LazyClassInitializer init, Source source) {
        super(ts, init, source);
        okAsType = false;
        cases = new HashSet();
    }

    public void parent(ExtensionType et) {
        parent = et;
    }

    public ExtensionType parent() {
        return parent;
    }

    public boolean okAsType() {
        return okAsType;
    }

    public ExtensionType okAsType(boolean b) {
        ExtensionType_c r = (ExtensionType_c) copy();
        r.okAsType = b;
        return r;
    }

    protected ParsedClassType interfaceType;

    public ParsedClassType interfaceType() {
        if (interfaceType == null) {
            try {
                interfaceType = (ParsedClassType) clone();
                interfaceType.superType(null);
                interfaceType.flags(interfaceType.flags().Interface());
            } catch (CloneNotSupportedException e) {
                interfaceType = null;
            }
        }
        return interfaceType;
    }

    public String extensionName() {
        return extensionName;
    }

    public void extensionName(String extensionName) {
        this.extensionName = extensionName;
    }

    public boolean isTop() {
        return isTop;
    }

    public void isTop(boolean isTop) {
        this.isTop = isTop;
    }

    public Set cases() {
        return cases;
    }

    public void addCase(Type t) {
        cases.add(t);
    }

    public boolean implicitlyApplicable(Type t) {
        // Only apply extensions to arguments which haven't been extended
        // explicitly.
        if (t instanceof ExtensionType || t instanceof UnextendedType) 
            return false;
        
        return applicableTo(t);
        
    }

    public boolean applicableTo(Type t) {
        EJavaTypeSystem ts = (EJavaTypeSystem) typeSystem();
        boolean oldexp = ts.getUseExpanders();
        ts.setUseExpanders(false);
            //now implement the actual check. this should involve type inference when we get to generics.
            boolean r = ts.isSubtype(t, this.superType());
        ts.setUseExpanders(oldexp);
        return r;
    }
}
