package polyglot.ext.ejava.visit;

import polyglot.ast.Call;
import polyglot.ast.Node;
import polyglot.ast.NodeFactory;
import polyglot.ast.Special;
import polyglot.ext.ejava.types.ExtensionType;
import polyglot.frontend.Job;
import polyglot.types.ClassType;
import polyglot.types.MethodInstance;
import polyglot.types.SemanticException;
import polyglot.types.TypeSystem;
import polyglot.visit.ContextVisitor;
import polyglot.visit.NodeVisitor;

/**
 * Visitor that makes sure extension's body does not: override any of its base
 * class's methods. access any methods of the base class that would result in
 * javac errors in the generated code
 */
public class ExtensionBodyRestrictionEnforcer extends ContextVisitor {
    public ExtensionBodyRestrictionEnforcer(Job job, TypeSystem ts, NodeFactory nf) {
        super(job, ts, nf);
    }

    public Node leaveCall(Node old, Node n, NodeVisitor v) throws SemanticException {
        if (n instanceof Call) {
            Call cn = (Call) n;

            ClassType currClass = context().currentClass();
            if (currClass instanceof ExtensionType && cn.target() instanceof Special) {
                Special target = (Special) cn.target();
                MethodInstance mi = cn.methodInstance();
                // TODO: forbid "super.m()" if m is a method of the expander's
                // base type
                if ((target.kind().equals(Special.SUPER) || !currClass.hasMethodImpl(mi))
                        && mi.flags().isProtected()) {
                    polyglot.types.Package extensionPackage = currClass.package_(), instancePackage = ((ClassType) mi.container()).package_();
                    if (!(extensionPackage == null && instancePackage == null || extensionPackage != null
                            && extensionPackage.equals(instancePackage)))
                        throw new SemanticException("this method cannot be accessed from "
                                + "an expander", cn.position());
                }
            }
            return super.leaveCall(old, n, v);
        }
        return super.leaveCall(old, n, v);
    }
}
